#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <math.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <stdint.h>

#include <sys/times.h>
#include <sys/time.h>
#include <sys/types.h>


#include "vec.h"
#include "sathys.h"
#include "analyseRL.h"
#include "heap.h"
#include "metaheuristique.h"


/************************************/
/* Compilation flags                */
/***********************************/

#define LUBY 1
#define TIRETS 141
#define INIT_LUBY 64

/* WSAT variable*/
int nbRestartTotal = 0;
BIGINT numflip;		        /* number of changes so far */
int maxSat = BIG;
int nbAlea = 0;
int *breakcount;	        /* number of clauses that become unsat if var if flipped */
int *makecount; 	        /* number of clauses that become sat if var if flipped */

double numCallCDCL = 0;

int nbAugmentation = 0;
int freeFlip = 0;

int *bestInterpretation;
int *variableAugmentation;
unsigned int *varEqui;

int numfalseBin;

int maxSatNoBinaire;
int maxSatTmp = 0;

unsigned int *falseBin;                /* falseBin[0] = nb lit false bin */
int *whereFalseBin, *nbLitFalseBin;


/* No longer using constant CLK_TCK - it is deprecated! */
/* Instead: */
long ticks_per_second;

/**************************************/
/* CDCL variable                      */
/**************************************/

extern double weightUp;
extern int *posChoice;
extern unsigned int **reason;
extern int *partial_assign;
extern int *isAssign;
extern int decisionLevel;
extern int tail;
extern int *level;
extern int *vTemp;
extern int indVtmp;
extern vector *watched;


/************************************/
/* Main data structures             */
/************************************/

int numatom = 0;
int numclause = 0;
int numliterals;

int nbRemoveRed = 0;
int nbRemoveRedByBin = 0;
int nbRemoveInit = 0;
int nbRemoveLearnt = 0;

int nbFlip = 0;
int whenLearnt = 2;

int testFalseLeartn = 0;
unsigned int ** clause = NULL;		/* clauses to be satisfied */
				/* indexed as clause[clause_num][literal_num] */
int *atom;		        /* value of each atom */

int resultat = OUT;
int nbClauseAdd = 0;
int nbClauseDer = 0;
int nbConflitDer = 0;
int maxAtom = 100000;
int nbClauseBin = 0;

int nbConf = 0;
int nbCleaning = 0;
double bornconf = 100;

double cla_inc = 1.0, clause_decay = 1 / 0.999;
double *weightVar = NULL;

int numtry = 0;			/* total attempts at solutions */
int nbRepetition = 0;

/* Randomization */
unsigned int seed;  /* Sometimes defined as an unsigned long int */

struct timeval tv;
struct timezone tzp;

/* Statistics */

double expertime;
double totalTime=0;

int nbPush = 0;


/* Learnt clause */
int numclauseLearnt = 0;
unsigned int ** clauseLearnt = NULL;

/* binary clause */
binaryStruct *binClause;

/************************************/
/* Forward declarations             */
/************************************/

void parse_parameters(int argc,char *argv[]);
double elapsed_seconds(void);
void initprob(void); 
void handle_interrupt(int sig);
long super(int i);
void print_statistics_header(void);
void showInter();
void showFinal();
void removeClauseDB(int cl);
void cleaningDB();
void solve(int nbConflict);

double sizeLearnt;

int nbPC =0;
int nbCf = 0;
BIGINT nbPropagation = 0;

int var1 = 0;
int stat;

int nbCauseBeforeReduce;
int posSimplify = 0;

enum heuristics { BEST, NOVELTY, RNOVELTY, NOVELTY_PLUS, RNOVELTY_PLUS};
char * heuristic_names[] = { "best", "novelty", "rnovelty", "novelty+", "rnovelty+"};
static int (*pickcode[])(void (*)()) = {bestHyb, noveltyHyb, rnoveltyHyb, noveltyPlusHyb, rnoveltyPlusHyb};
int heuristic = BEST;



/***************************************************************************************/
/* Memory used     *********************************************************************/
/***************************************************************************************/

int memReadStat(int field)
{
  char name[256];
  pid_t pid = getpid();
  int     value;
  sprintf(name, "/proc/%d/statm", pid);
  FILE*   in = fopen(name, "rb");
  if (in == NULL) return 0;
  for (; field >= 0; field--)
    assert(fscanf(in, "%d", &value));
  fclose(in);
  return value;
}

uint64_t memUsed() 
{
  return (uint64_t)memReadStat(0) * (uint64_t) getpagesize();
}

/*****************************************************************************************/
/* Main           ************************************************************************/
/*****************************************************************************************/

#define NBCLAUSEREDUCE 1000
#define NBADDREDUCE 2000

int main(int argc,char *argv[])
{
  ticks_per_second = sysconf(_SC_CLK_TCK);
  gettimeofday(&tv,&tzp);
  seed = (unsigned int)((( tv.tv_sec & 0177 ) * 1000000) + tv.tv_usec);
  parse_parameters(argc, argv);
  srandom(seed);

  signal(SIGINT, handle_interrupt);
  signal(SIGTERM, handle_interrupt);
  signal(SIGALRM, handle_interrupt);

  initprob();
  
  print_statistics_header();

  (void) elapsed_seconds();
  
  nbClauseDer = numclause;

  sizeLearnt = NBCLAUSEREDUCE;
  bornconf = 1;

  initWSAT();
  wsat();
  termineProg(OUT,"UNKNOWN");
  return 0;
}/* main */

/**
   This function scan the next parameters (int).
 */
void scanone(int argc, char *argv[], int i, int *varptr)
{
    if (i>=argc || sscanf(argv[i],"%i", varptr)!=1){
      fprintf(stderr, "c Bad argument %s\n", i<argc ? argv[i] : argv[argc-1]);
      exit(-1);
    }
}/* scanone */

/**
   This function scan the next parameters (BIGINT).
 */
void scanonell(int argc, char *argv[], int i, BIGINT *varptr)
{
  if (i>=argc || sscanf(argv[i],"%Li",varptr)!=1){
    fprintf(stderr, "c Bad argument %s\n", i<argc ? argv[i] : argv[argc-1]);
	exit(-1);
  }
}/* scanonell */


/**
   Parse parameters
 */
void parse_parameters(int argc,char *argv[])
{
  int i;
  int temp;
  
  for (i=1 ; i < argc ; i++)
    {
      if (strcmp(argv[i],"-stat") == 0)
	stat = 1;
      else if (strcmp(argv[i],"-seed") == 0)
	{
	  scanone(argc,argv,++i, &temp);
	  seed = (unsigned int)temp;
	}
      else if (strcmp(argv[i],"-p") == 0)
	{
	  scanone(argc,argv,++i,&temp);
	  numCallCDCL = temp;
	}else if (strcmp(argv[i],"-best") == 0) heuristic = BEST;
      else if (strcmp(argv[i],"-novelty") == 0) heuristic = NOVELTY;
      else if (strcmp(argv[i],"-rnovelty") == 0) heuristic = RNOVELTY;      
      else if (strcmp(argv[i],"-novelty+") == 0) heuristic = NOVELTY_PLUS;      
      else if (strcmp(argv[i],"-rnovelty+") == 0) heuristic = RNOVELTY_PLUS;      
    }
}/* parse_parameters */

/**
   Luby
 */
long super(int i)
{
    long power;
    int k;

    if (i<=0){
      fprintf(stderr, "c bad argument super(%d)\n", i);
      exit(1);
    }
    /* let 2^k be the least power of 2 >= (i+1) */
    k = 1;
    power = 2;
    while (power < (i+1)){
      k += 1;
      power *= 2;
    }
    if (power == (i+1)) return (power/2);
    return (super(i - (power/2) + 1));
}/* super */

/**
   Handle interrupt
*/
void handle_interrupt(int sig)
{
  showFinal();
  exit(0);
}/* handle_interrupt */

/**
   Compare variable
 */
static int cmpInt(const void *p1, const void *p2)
{
  int a = *(int*)p1;
  int b = *(int*)p2;
  if(a<0) a = -a;
  if(b<0) b = -b;
  return a - b;
}//cmpInt

/**
   Compare integer inverse
 */
static int cmpEntInv(const void *p1, const void *p2)
{
  int a = *(int*)p1;
  int b = *(int*)p2;
  
  return b - a;
}


int nbEqui = 0;
int nbPU = 0;


/**
   Libre l'ensemble des clauses binaires avec une certaine variable
*/
void freeBinary(unsigned int var)
{
  if(binClause[var].nbBin)
    {
      free(binClause[var].tabBin);
      binClause[var].tabBin = NULL;
      binClause[var].nbBin = 0;
    }
  
}/* freeBinary */
  
/**
   Init prob : parse problem
*/
void initprob(void)
{
  int i;
  int j = 0;
  unsigned int stp1;
  unsigned int stp2;
  int lastc;
  int nextc;
  int storeptr[MAXLENGTH];
  int freestore;
  int lit;
  int tau;
  int sz;
  
  while ((lastc = getchar()) == 'c' || lastc == '\n')
    {
      while ((nextc = getchar()) != EOF && nextc != '\n');
    }
  ungetc(lastc,stdin);
  if (scanf("p cnf %i %i",&numatom, &numclause) != 2)
    {
      fprintf(stderr,"c initprob 1 : Bad input file\n");
      exit(-1);
    }

  numclause += 10;
  maxAtom = numatom + 2;
  while (getchar()!='\n');

  /* initMyVariable */  
  if(!(atom = (int *) calloc(maxAtom + 1, sizeof(int)))){perror("atom"); exit(1);}  
  if(!(bestInterpretation = (int *) calloc(maxAtom + 1, sizeof(int)))){perror("atom"); exit(1);}  
  if(!(variableAugmentation = (int *) calloc(maxAtom + 1, sizeof(int)))){perror("atom"); exit(1);}
  if(!(varEqui = (unsigned int *) calloc(maxAtom + 1, sizeof(unsigned int)))){perror("atom"); exit(1);}
  
  initDPLL();
  
  /* on prpare pour les clauses binaires */
  binClause = (binaryStruct *) malloc(sizeof(binaryStruct) * ((maxAtom + 1) * 2));
  for( i = 0 ; i < (2 * (maxAtom + 1)) ; i++)
    {
      binClause[i].res[0] = 2;
      binClause[i].res[1] = 0; /* une valeur au pif qui sera utilis pour l'analyse de conflit */
      binClause[i].res[2] = i; /* le littral qui sera la raison de la propagation */
      binClause[i].res[3] = -1; /* permet de dire que la clause n'est pas une clause apprise */
      binClause[i].nbBin = 0; /* le nombre de 'clause binaire' associ */
      binClause[i].tabBin = NULL;
    }

  freestore = 0;
  numliterals = 0;
  for(i = 0 ; i < numclause ; i++)
    {
      sz = -1;
      tau=0;
      do
        {
          sz++;
	  j = scanf("%d", &lit);
          if(j != 1)
	    {
	      lit = 0;
	      
	      if(j == EOF)
		{
		  numclause = i + 1;
		  if(!(sz))
		    {
		      numclause--;
		      goto end_loop;
		    }
		}
	    }
          
          /* show lit exist to clause[i] or it's a tauto */
	  if(lit)
	    {
	      for(j=0 ; j<(sz) ; j++)
		if((storeptr[j]==lit) || (storeptr[j]==-lit)) break;
	      if(j < sz)
		{
		  if(storeptr[j] == lit)
		    {
		      sz--;
		      continue;
		    }/*redundancy*/
		  else tau = 1;/*tautologie*/
		}
	    }
          
          if(lit != 0)
            {
              storeptr[sz] = lit; /* clause[i][size[i]] = j; */
              freestore--;
              numliterals++;
            }
        }
      while(lit != 0);
      assert(sz);
      
      if(tau)
        {
          numclause--;
          i--;
        }
      else
        {
	  if(sz == 1)
	    {
	      stp1 = (storeptr[0] > 0) ? storeptr[0]<<1 :  (-storeptr[0]<<1) + 1;
	      if(LIT_FALSE(stp1, isAssign[ABS(stp1)])) termineProg(UNS, "UNSAT");
	      if(isAssign[ABS(stp1)] == NOT_ASS)
		uncheckedEnqueue(stp1, NULL);
	      numclause--;
	      i--;
	    }else
	    {
	      if(sz == 2)
		{
		  stp1 = (storeptr[0] > 0) ? storeptr[0]<<1 :  (-storeptr[0]<<1) + 1;
		  stp2 = (storeptr[1] > 0) ? storeptr[1]<<1 :  (-storeptr[1]<<1) + 1;
		  
		  nbClauseBin++;		  
		  binClause[stp1].tabBin = (unsigned int*) realloc(binClause[stp1].tabBin,sizeof(unsigned int)*(binClause[stp1].nbBin+1));
		  binClause[stp1].tabBin[binClause[stp1].nbBin++] = stp2;
		  
		  binClause[stp2].tabBin = (unsigned int*) realloc(binClause[stp2].tabBin,sizeof(unsigned int)*(binClause[stp2].nbBin+1));
		  binClause[stp2].tabBin[binClause[stp2].nbBin++] = stp1;		 

		  numclause--;
		  i--;
		}else
		{
		  if(!(clause = (unsigned int **) realloc(clause, sizeof(unsigned int *) * (i + 1)))){perror("clause"); exit(1);}
		  if(!(clause[i] = (unsigned int *) malloc(sizeof(unsigned int)*(sz + 3))))
		    {
		      perror("size");
		      exit(1);
		    }  
		  qsort(storeptr, sz, sizeof(int), cmpInt);
		  for(j = 1 ; j <= sz ; j++)
		    {
		      if(storeptr[j -1] > 0)
			clause[i][j] = storeptr[j - 1]<<1;
		      else clause[i][j] = ((-storeptr[j - 1])<<1) + 1;
		    }
		  *clause[i] = sz;
		  clause[i][sz + 1] = -(i + 1);
		  clause[i][sz + 2] = 1;	
		  attachClause(clause[i]);
		}
	    }
	}
    }
 end_loop:  

  if(propagate()) termineProg(UNS, "UNSAT");
  posChoice[0] = tail;
  posSimplify = 0;
  initHeap();

  for(i = 1 ; i <= numatom ;i++) addHeap(i);
}/* initprob */


/**
   elapsed seconds
 */
double elapsed_seconds(void) 
{ 
   double answer;   
   answer = (double) clock() / CLOCKS_PER_SEC - totalTime;
   totalTime = (double) clock() / CLOCKS_PER_SEC;
   return answer; 
}/* elapsed_seconds */


/**********************************************************************************/
/**********************************************************************************/
/**********************************************************************************/

/**
   Print statistics header
 */
void print_statistics_header(void)
{
  int i;
  fprintf(stderr, "c \nc seed = %i, numatom = %i, numclause = %i, numliterals = %i\n",seed, numatom,numclause,numliterals);

  fprintf(stderr, "c ");
  for(i = 0 ; i< TIRETS ; fprintf(stderr, "-"), i++); fprintf(stderr, "\n");
  fprintf(stderr, "c | %6s | %7s | %7s | %8s | %6s | %7s | %9s | %7s |",
	  "resCDCL", "tps ", "#impl", "#learnt", "#conf", "#clean", "#propa", "#deci");
  fprintf(stderr, "| %6s | %7s | %15s | %7s | %7s |\n",
	  "tryLS", "numflip", " maxSat/maxNoBi", "nbAlea", "Equi");
  fprintf(stderr, "c ");
  for(i = 0 ; i< TIRETS ; fprintf(stderr, "-"), i++); fprintf(stderr, "\n");
}/* print_statistics_header */


/**
   Fonction permettant d'afficher les informations intermdiaire
 */
void showInter()
{
  double temps = elapsed_seconds();

  fprintf(stderr,
  "c | %7d | %7.1lf | %7d | %8d | %6d | %7d | %9lld | %7d |",
	  nbRestartTotal, temps, posChoice[0], numclauseLearnt, nbConf, nbCleaning, nbPropagation, nbPC);

  fprintf(stderr, "| %6d | %7lld | %7d/%7d | %7d | %7d |\n",
	  numtry, numflip, maxSatTmp, maxSatNoBinaire, nbAlea, nbEqui);
}/* showInter */

/**
   Fonction permettant de savoir le temps pass depuis le dbut du programme
   Retourne :
      - le temps pass depuis le dbut du programme
 */
double tps_ecouler()
{
   double answer;
   answer = (double) clock() / CLOCKS_PER_SEC;
   return answer; 
}//tps_ecouler


/**
   Fonction permettant d'afficher les rsultats du test
 */
void showFinal()
{
  int i;

  showInter();
  fprintf(stderr, "c ");
  for(i = 0 ; i< TIRETS ; printf("="), i++); printf("\n");
  fprintf(stderr, "c \nc -----------------------------------\n");
  fprintf(stderr, "c # temps = %lf\n", tps_ecouler());  
  fprintf(stderr, "c # cleaningDB = %d\n",nbCleaning);
  fprintf(stderr, "c # clause binaire = %d\n",nbClauseBin);
  fprintf(stderr, "c # literal removed by litRedundant = %d\n",nbRemoveRed);
  fprintf(stderr, "c # literal removed by litRedundantBin = %d\n",nbRemoveRedByBin);
  fprintf(stderr, "c # init clause removed by simplify = %d\n",nbRemoveInit);
  fprintf(stderr, "c # learnt clause removed by simplify = %d\n",nbRemoveLearnt);
  fprintf(stderr, "c # une clause fausse tait une apprise = %d\n", testFalseLeartn);
  fprintf(stderr, "c # equivalence dectecte = %d\n", nbEqui);
  uint64_t mem_used = memUsed();
  if(mem_used != 0) fprintf(stderr,"c # Memory used           : %.2f MB\nc \n", mem_used / 1048576.0);

  fprintf(stderr, "c ########################\n");
  fprintf(stderr, "c ####   ##  ###   ## ####\n");
  fprintf(stderr, "c #### #### # ## #### ####\n");
  fprintf(stderr, "c ####   ##  ###   ##  ###\n");
  fprintf(stderr, "c ########################\n");  
  fprintf(stderr, "c # restart = %d\n",nbRestartTotal);
  fprintf(stderr, "c # conflit = %d\n",nbCf);
  fprintf(stderr, "c # decision = %d\n",nbPC);
  fprintf(stderr, "c # propagation = %lld \t (%lf (prop/sec))\nc \n",nbPropagation, nbPropagation / tps_ecouler());

  fprintf(stderr, "c ########################\n");  
  fprintf(stderr, "c ### ### #   ##   ##   ##\n");  
  fprintf(stderr, "c ### # # # #### # ### ###\n");  
  fprintf(stderr, "c ###  #  ### ## # ### ###\n");  
  fprintf(stderr, "c ########################\n");  
  fprintf(stderr, "c # numtry = %d\n", numtry);
  fprintf(stderr, "c # numflips = %lld\n", numflip);
  fprintf(stderr, "c # maxSat = %d\n", maxSat);
  fprintf(stderr, "c # nbAugmentation = %d\n", nbAugmentation);
  fprintf(stderr, "c # freeFlip = %d\n", freeFlip);

  fprintf(stderr, "c \n");


  switch(resultat)
    {
    case SAT :
      {
        printf("s SATISFIABLE\n");
        printf("v ");
        for(i = 1 ; i<(numatom + 1); i++)
          if(atom[i]) printf("%d ", i); else printf("%d ",-i);
        printf("0\n");
        exit(10);
        break;
      }
    case UNS :
      {
        printf("s UNSATISFIABLE\n");
        exit(20);
        break;
      }
    case OUT : 
      {
        printf("s UNKNOWN\n");
      }
    }  
}//showFinal


/**
   Fonction permettant de grer la terminaison du programme
   En entre : 
      - res, dans quel tat le programme c'est termin
      - comment, une chaine de caractère dcrivant l'tat de sorti
 */
void termineProg(int res, char *comment)
{
  resultat = res;  
  kill(getpid(), SIGTERM);
}/* termineProg */


/**
   Realloc memory variable
 */
void realloue()
{
  if(!(clauseLearnt = (unsigned int **) realloc(clauseLearnt, sizeof(unsigned int *)*(((numclauseLearnt/MAXCLAUSEBLOC)+1)*MAXCLAUSEBLOC))))
    {
      perror("clause\n");
      exit(1);
    }
}/* realloue */


/**
   Add new clause of data base
   In entry :
       - c, one table (malloc)       
   Return : This function return 0 if the new clause is added, otherwise >0

 */
int addClauseDB(unsigned int *c)
{
  nbClauseAdd++;

  if(*c == 2)
    {       
      nbClauseBin++;
      
      binClause[c[1]].tabBin = (unsigned int *) realloc(binClause[c[1]].tabBin, sizeof(unsigned int) * (binClause[c[1]].nbBin + 1));
      binClause[c[1]].tabBin[binClause[c[1]].nbBin++] = c[2];
      
      binClause[c[2]].tabBin = (unsigned int *) realloc(binClause[c[2]].tabBin, sizeof(unsigned int) * (binClause[c[2]].nbBin + 1));
      binClause[c[2]].tabBin[binClause[c[2]].nbBin++] = c[1];
      
      if(WASS_TRUE(c[1], atom[ABS(c[1])]) && WASS_FALSE(c[2], atom[ABS(c[2])])) breakcount[ABS(c[1])]++;
      if(WASS_FALSE(c[1], atom[ABS(c[1])]) && WASS_TRUE(c[2], atom[ABS(c[2])])) breakcount[ABS(c[2])]++;
      if(WASS_FALSE(c[1], atom[ABS(c[1])]) && WASS_FALSE(c[2], atom[ABS(c[2])]))
	{
	  makecount[ABS(c[1])]++;
	  makecount[ABS(c[2])]++;
	  if(!(nbLitFalseBin[ABS(c[1])]++)) addLitBinFalse(c[1]);
	  if(!(nbLitFalseBin[ABS(c[2])]++)) addLitBinFalse(c[2]);
	  numfalseBin++;
	}
    }else
    {
      if(!(numclauseLearnt % MAXCLAUSEBLOC))
	realloue();
      
      clauseLearnt[numclauseLearnt] = c;
      attachClause(c);
      ++numclauseLearnt;
    }

  return 1;
}/* addClauseDB */


/**
   Remove clause learnt
   In entry :
        - cl, index of the clause to remove
 */
void removeClauseInit(unsigned int *cl)
{
  detachClause(cl);  
  free(cl);
}/* removeClause */

/**
   Remove clause learnt
   In entry :
        - cl, index of the clause to remove
 */
void removeClause(unsigned int *cl)
{
  detachClause(cl);
  free(cl);
}/* removeClause */


/**
   Coompare
 */
static int cmpWeightCl(const void *p1, const void *p2)
{
  unsigned int *c1 = *(unsigned int **) p1;
  unsigned int *c2 = *(unsigned int **) p2;

  return c2[*c2 + 1] - c1[*c1 + 1];
}//cmpWeightCl

int locked(unsigned int * cl){ return (reason[ABS(cl[1])] == cl) && (LIT_TRUE(cl[1], isAssign[ABS(cl[1])])) && (level[ABS(cl[1])]);}
int trueClause(unsigned int *cl){ return (LIT_TRUE(cl[1], isAssign[ABS(cl[1])])) && (!level[ABS(cl[1])]); }

/**
   Refresh data base
 */
void cleaningDB()
{
  int i, j = 0; 
  unsigned int *cl;
  
  for( i = 0 ; i<numclauseLearnt ; i++)
    {
      cl = clauseLearnt[i];
      
      cl[*cl + 1] = 0;
      for(j = 1 ; j<=*cl ; j++)
	{
	  if(1 & (bestInterpretation[ABS(cl[j])] ^ cl[j]))
	    cl[*cl + 1]++;
	}
    }

  qsort(clauseLearnt, numclauseLearnt, sizeof(unsigned int *), cmpWeightCl);  

  for(i = j = 0 ; (i < (numclauseLearnt / 2)) && (clauseLearnt[i][*clauseLearnt[i] + 1] > 2) ; i++)
    {
      if (!locked(clauseLearnt[i]))
	removeClause(clauseLearnt[i]);
      else clauseLearnt[j++] = clauseLearnt[i];
    }

  for( ; i<numclauseLearnt ; i++)
    {
      if(trueClause(clauseLearnt[i]))
	removeClause(clauseLearnt[i]);
      else clauseLearnt[j++] = clauseLearnt[i];
    }
  
  numclauseLearnt = j;
  sizeLearnt += NBADDREDUCE;
}/* cleaningDB */


/**
   Return the next decision variable
 */
unsigned int pickBranchLit()
{
  unsigned var;
  int i;
  do
    {
      var = nextOrderHeap();
    }
  while(isAssign[var] != NOT_ASS && var);
  
  assert((isAssign[var] == NOT_ASS) || !var);
  
  if(!var)
    {
      for( i = 1 ; i <= numatom; i++) assert(isAssign[i] == NOT_ASS);
      for(i = 0 ; i<tail ; i++)             
	{
	  var = partial_assign[i];
	  atom[ABS(var)] = (SIGN(var))? 0 : 1;
	}
      termineProg(SAT,"ASSIGNMENTS FOUND");
    }
  
  return (var<<1) | ((bestInterpretation[var] + 1) & 1);    
}/* pickBranchLit */

/**********************************************************************************/
/************************* **** ***    ****     ***     ***************************/
/************************* **** *** ******* *** ***** *****************************/
/************************* *  * ***    ****     ***** *****************************/
/*************************  **  ******* *** *** ***** *****************************/
/************************* **** ***     *** *** ***** *****************************/
/**********************************************************************************/

int **occurence;                /* where each literal occurs */
				/* indexed as occurence[literal+MAXATOM][occurence_num] */
int *numoccurence;       	/* number of times each literal occurs */
int * numtruelit;		/* number of true literals in each clause */
int * false;			/* clauses which are false */
int * wherefalse;		/* where each clause is listed in false */

int *watch1;
int *watch2;

int *changed;    		/* step at which atom was last flipped */

unsigned int *bestAtom;
int nbConflict;


BIGINT numnullflip;		/*  number of times a clause was picked, but no  */

int numfalse;			/* number of false clauses */
int lastNumflipRestart = 0;


/**
   Fonction permettant de supprimer les clauses qui sont satisfaite
 */
void simplify()
{
  int i, j, k, indCl;
  unsigned int var;
  unsigned int *cl;

  int *pi, *pj, *end;

  if(decisionLevel != 0) return;

  for( ; posSimplify < posChoice[0] ; posSimplify++)
    {
      var = partial_assign[posSimplify];
      freeBinary(var);
      freeBinary(NEG(var));

      /* on supprime les occurences */
      for( i = 0 ; i<numoccurence[var] ; i++)
	{
	  indCl = occurence[var][i];
	  cl = clause[indCl];
	  
	  for( j = 1 ; j <= *cl ; j++)
	    {
	      if(cl[j] == var) continue;
  
	      /* on  supprime dans occurence */
	      for(pi = pj = occurence[cl[j]], end = &(occurence[cl[j]][numoccurence[cl[j]]]) ; pi != end; pi++)
		if(*pi != indCl) *(pj++) = *(pi);	      
	      numoccurence[cl[j]]--;
	    }
	}
      
      /* on supprime la clause */
      qsort(occurence[var], numoccurence[var], sizeof(int), cmpEntInv);

      for( pi = occurence[var], end = &(occurence[var][numoccurence[var]]) ; pi != end ; )
	{
	  indCl = *(pi++);
	  cl = clause[indCl];
	  assert(numtruelit[indCl]);
	  
	  nbRemoveInit++;
	  detachClause(cl);
	  free(cl);

	  if(indCl < (numclause - 1))
	    {
	      cl = (clause[indCl] = clause[--numclause]);

	      for( j = 1 ; j <= *cl ; j++)
		{
		  assert(cl[j] != var);
		  
		  for( k = 0 ; (k < numoccurence[cl[j]]) && (occurence[cl[j]][k] != numclause) ; k++);
		  assert(k < numoccurence[cl[j]]);
		  occurence[cl[j]][k] = indCl;
		}
	      
	    }else numclause--;
	}
      
      if(numoccurence[var])
	{
	  free(occurence[var]);
	  occurence[var] = NULL;
	  numoccurence[var] = 0;
	}
    }
}/* simplify */


/**
   Initialisation de la partie Walksat
 */
void initWSAT()
{
  int i, j;
  unsigned int *cl;

  if(!(numtruelit = (int *) malloc(sizeof(int) * numclause))){perror("numtruelit"); exit(1);}
  if(!(false = (int *) malloc(sizeof(int) * numclause))){perror("false"); exit(1);}
  if(!(wherefalse=(int *)malloc(sizeof(int) * numclause))){perror("wherefalse"); exit(1);}
  if(!(watch1 = (int *) malloc(sizeof(int) * numclause))){perror("watch1"); exit(1);}
  if(!(watch2 = (int *) malloc(sizeof(int) * numclause))){perror("watch2"); exit(1);}
  if(!(changed = (int *)  calloc(numatom + 1, sizeof(int)))){perror("initMyVariable");exit(1);}
  
  if(!(numoccurence = (int *) calloc(2*(numatom + 1), sizeof(int)))){perror("numoccurence"); exit(1);}
  if(!(occurence = (int **) calloc(2*(numatom + 1), sizeof(int *)))){perror("occurence"); exit(1);}
  if(!(atom = (int *) calloc(numatom + 1, sizeof(int)))){perror("atom"); exit(1);}
  if(!(breakcount = (int *) calloc(numatom + 1, sizeof(int)))){perror("breakcount"); exit(1);}

  if(!(makecount= (int *) calloc(numatom + 1, sizeof(int)))){perror("makecount"); exit(1);}
  if(!(bestAtom = (unsigned int *) calloc(numatom + 1, sizeof(unsigned int)))){perror("best"); exit(1);}  

  if(!(falseBin = (unsigned int *) calloc(maxAtom + 1, sizeof(unsigned int)))){perror("atom"); exit(1);}
  if(!(whereFalseBin = (int *) calloc(maxAtom + 1, sizeof(int)))){perror("atom"); exit(1);}
  if(!(nbLitFalseBin = (int *) calloc(maxAtom + 1, sizeof(int)))){perror("atom"); exit(1);}

  for(i = 0 ; i < 2*(numatom + 1) ; i++)
    {
      occurence[i] = NULL;
      numoccurence[i] = 0;
    }
  
  for(i = 0 ; i < numclause ; i++)
    {
      cl = clause[i];
      for(j = 1 ; j <= *cl ; j++)
        {          
          if(!(numoccurence[cl[j]] % BLOC))
            if(!(occurence[cl[j]]= (int *) realloc(occurence[cl[j]],(sizeof(int)*(numoccurence[cl[j]]+BLOC)))))
              {
                fprintf(stderr,"initprob : ERROR - allocation memory for occurence\n");
                exit(-1);
              }
          
          occurence[cl[j]][numoccurence[cl[j]]] = i;
          numoccurence[cl[j]]++;
        }
    }
}/* initWSAT */


/**
   Init l'interprtation complte
 */
void initInterpretation(int flag)
{
  int i, j;
  unsigned int lit;

  if(flag != 1)
    {
      switch(flag)
	{
	case 0 :
	  {
	    for(i = 1 ; i <= numatom ; i++)
	      {
		j = numoccurence[i << 1 ] - numoccurence[(i << 1) | 1]
		  + binClause[i << 1].nbBin - binClause[(i << 1) | 1].nbBin;
		atom[i] = (j > 10) ? 1 : ((j <= 10) ? 0 :random() % 2);
	      }
	    break;
	  }
	case 2 :
	  {	    
	    nbAugmentation = 0;
	    for( i = 1 ; i <= numatom ; i++)
	      {
		nbAugmentation += (variableAugmentation[i] >>= 1);
		atom[i] = bestInterpretation[i];	    
	      }
	    break;	
	  }
	}
      for( i = 0 ; i<tail ; i++)
	{      
	  lit = partial_assign[i];
	  atom[ABS(lit)] = (lit & 1) ^ 1;
	}
    }
  
}/* initInterpretation */

/**
   Init les clauses de la base
 */
void initClauseBase()
{
  int i, j;
  int thetruelit = 0;
  int truelit1;
  int truelit2;
  unsigned int *cl;

  numfalse = 0;

  /* Initialize breakcount and makecount in the following: */
  for(i = 0 ; i < numclause ; i++)
    {      
      cl = clause[i];
      assert(cl);
      
      numtruelit[i] = 0;
      truelit1 = 0;
      truelit2 = 0;
      for(j = 1 ; j <= *cl ; j++)
	{
	  if(WASS_TRUE(cl[j], atom[ABS(cl[j])]))
	    {
	      numtruelit[i]++;
	      thetruelit = cl[j];
	      if (!truelit1) truelit1 = cl[j];
	      else if (truelit1 && !truelit2) truelit2 = cl[j];
	    }
	}
      
      if(!numtruelit[i])
	{
	  wherefalse[i] = numfalse;
	  false[numfalse] = i;	  
	  numfalse++;
	  for(j = 1 ; j <= *cl ; j++) makecount[ABS(cl[j])]++;	  
	}
      else if (EQ(numtruelit[i], 1))
	{
	  breakcount[ABS(truelit1)]++;
	  watch1[i] = ABS(truelit1);
	}
      else /*if (numtruelit[i] == 2)*/
	{
	  watch1[i] = ABS(truelit1);
	  watch2[i] = ABS(truelit2);
	}
    }
}/* initClauseBase */


/**
   Init les clauses binaire
 */
void initBinClause()
{
  int i, j;
  unsigned int var, lit;

  *falseBin = 0;  
  numfalseBin = 0;
  
  for( i = 1 ; i <= numatom ; i++)
    {
      nbLitFalseBin[i] = 0;
      lit = (i<<1) | (atom[i] & 1);

      for( j = 0 ; j<binClause[lit].nbBin ; j++)
	{	  
	  var = binClause[lit].tabBin[j];
	  if(WASS_FALSE(var, atom[ABS(var)]))
	    {
	      makecount[ABS(var)]++;
	      numfalseBin++;
	      nbLitFalseBin[i]++;
	    }
	  else breakcount[ABS(var)]++;
	}
      
      if(nbLitFalseBin[i]) addLitBinFalse(lit);	
    }
  
  numfalseBin >>= 1;

}/* initBinClause */

/**
   Initialyze : breakcount, makecount, ....
 */
void init(int flag)
{
  int i;

  for(i = 1 ; i <= numatom ; i++)
    {
      changed[i] = -1;
      breakcount[i] = 0;
      makecount[i] = 0;
    }
  
  initInterpretation(flag);
  initBinClause();
  initClauseBase();
  
  maxSatTmp = ( maxSatNoBinaire = numfalse ) + numfalseBin;
}/* init */ 

/**
   Wsat call cleaningDB
 */
void callCleaningDB()
{
  nbCleaning++;
  cleaningDB();
  cancelUntil(0);
  if(posSimplify < posChoice[0]) simplify();
  init(2);  
  lastNumflipRestart = 0.1 * ((numatom < 5000) ? 5000 : numatom) * nbCleaning;
}/* callCleaningDB */

/**
   Supprime un littral du tableau falseBin
 */
void delLitBinFalse(unsigned int var)
{
  falseBin[whereFalseBin[ABS(var)]] = falseBin[(*falseBin)--];
  whereFalseBin[ABS(falseBin[whereFalseBin[ABS(var)]])] = whereFalseBin[ABS(var)];
}/* delLitBinFalse */

/**
   Ajoute un littral (tel que la variable n'est pas prsent) dans falseBin
 */
void addLitBinFalse(unsigned int var)
{
  falseBin[++(*falseBin)] = var;
  whereFalseBin[ABS(var)] = *falseBin;
}/* addLitBinFalse */


/*
  new flipping function based on SAT2004 submission work 
  Rend toflip vrai, c'est  dire que si il est positive et
  qu'il est affect  faux on le flip sinon on ne fait rien.
*/
void flipatom(unsigned int toflip)
{   
  int i, j;			
  int cli;
  int numocc;
  unsigned int absToflip = ABS(toflip);
  int * occptr;
  unsigned int *cl;
  unsigned int var;

  if(!toflip || WASS_TRUE(toflip, atom[ABS(toflip)]))
    {
      numnullflip++;
      return;
    }

  numflip++;
  
  atom[absToflip] = 1 ^ atom[absToflip];
  numocc = numfalseBin;
  
  for( i = 0 ; i<binClause[toflip].nbBin ; i++)
    {
      var = binClause[toflip].tabBin[i];
      if(WASS_TRUE(var, atom[ABS(var)]))
	breakcount[ABS(var)]--;
      else 
	{
	  makecount[absToflip]--;
	  makecount[ABS(var)]--;
	  breakcount[absToflip]++;
	  numfalseBin--;
	  if(!(--nbLitFalseBin[ABS(var)])) delLitBinFalse(var);
	}
    }

  numocc = nbLitFalseBin[absToflip];
  nbLitFalseBin[absToflip] = 0;

  /* on commence par les binaires */
  for( i = 0 ; i < binClause[NEG(toflip)].nbBin ; i++)
    {      
      var = binClause[NEG(toflip)].tabBin[i];
      if(WASS_TRUE(var, atom[ABS(var)]))
	breakcount[ABS(var)]++;
      else 
	{
	  makecount[absToflip]++;
	  makecount[ABS(var)]++;
	  breakcount[absToflip]--; 
	  numfalseBin++;
	  nbLitFalseBin[absToflip]++;
	  if(!(nbLitFalseBin[ABS(var)]++)) addLitBinFalse(var);	    
	}
    }

  if(numocc && nbLitFalseBin[absToflip]) falseBin[whereFalseBin[absToflip]] = NEG(toflip);
  else if(!numocc && nbLitFalseBin[absToflip]) addLitBinFalse(NEG(toflip));
  else if(numocc && (!nbLitFalseBin[absToflip])) delLitBinFalse(toflip); 
  
  /* les clauses de taille suprier  2 */
  toflip = NEG(toflip);
  numocc = numoccurence[toflip];
  occptr = occurence[toflip];
  changed[absToflip] = numflip;
  for(i = 0 ; i < numocc ;i++)
    {
      /* cli = occurence[MAXATOM-toenforce][i]; */
      cli = *(occptr++);
      
      if (--numtruelit[cli] == 0)
	{
	  false[numfalse] = cli;
	  wherefalse[cli] = numfalse;
	  
	  numfalse++;
	  /* Decrement toflip's breakcount */
	  breakcount[absToflip]--;
	  
	  /* Increment the makecount of all vars in the clause */
	  cl = clause[cli];	  
	  for (j = 1 ; j<=*cl ; j++)
	    makecount[ABS(cl[j])]++;	  
	}
      else if (numtruelit[cli] == 1)
	{
	  if(EQ(watch1[cli], absToflip))
	    {
	      assert(watch1[cli] != watch2[cli]);
	      watch1[cli] = watch2[cli];
	    }
	  breakcount[watch1[cli]]++;
	}
      else
	{ /* numtruelit[cli] >= 2 */
	  if(EQ(watch1[cli], absToflip))
	    {
	      /* find a true literal other than watch1[cli] and watch2[cli] */
	      cl = clause[cli];	  
	      for (j = 1 ; j<=*cl ; j++)
		{
		  var = cl[j];
		  if (WASS_TRUE(var,atom[ABS(var)]) && (DIF(ABS(var),watch1[cli])) && (DIF(ABS(var),watch2[cli])))
		    {
		      watch1[cli] = ABS(var);
		      break;
		    }
		}
	    }
	  else if (EQ(watch2[cli], absToflip)) 
	    {
	      /* find a true literal other than watch1[cli] and watch2[cli] */
	      cl = clause[cli];	  
	      for (j = 1 ; j <= *cl ; j++)
		{
		  var = cl[j];
		  if (WASS_TRUE(var,atom[ABS(var)]) && (DIF(ABS(var),watch1[cli])) && (DIF(ABS(var),watch2[cli])))
		    {
		      watch2[cli] = ABS(var);
		      break;
		    }
		}
	    }
	}
    }
  
  toflip = NEG(toflip);
  numocc = numoccurence[toflip];

  occptr = occurence[toflip];
  for(i = 0; i < numocc; i++)
    {      
      /* cli = occurence[MAXATOM+toenforce][i]; */
      cli = *(occptr++);

      if (++numtruelit[cli] == 1)
	{
	  numfalse--;
	  false[wherefalse[cli]] = false[numfalse];
	  wherefalse[false[numfalse]] = wherefalse[cli];
	  /* Increment toflip's breakcount */
	  breakcount[absToflip] ++;
	  
	  /* Decrement the makecount of all vars in the clause */
	  cl = clause[cli];      
	  for (j = 1; j<=*cl; j++) makecount[ABS(cl[j])]--;
	  watch1[cli] = absToflip;
	}
      else if (EQ(numtruelit[cli], 2))
	{
	  watch2[cli] = absToflip;
	  breakcount[watch1[cli]]--;
	}
    } 
}/* flipatom */


int compareWeightvar (const void * a, const void * b)
{
  unsigned int v1 = *(unsigned int*) a;
  unsigned int v2 = *(unsigned int*) b;
  return weightVar[ABS(v2)] - weightVar[ABS(v1)];
}


/**
   Fixe une variable de la formule
   In entry :
       - tofix, a variable
 */
void fixeVariable1(unsigned int var)
{
  unsigned int *clConf;
  unsigned int *cl;

  int tailTmp = tail;

  pushNewVariable(var);
  nbPC++;
  cl = propagate();
  
 classicalBacktrack:
  if (cl != NULL)
    {
      // CONFLICT
      nbConflict--;
      nbConf++;
      nbCf++;
      
      if(decisionLevel == 0) termineProg(UNS,"UNSATISFIABLE");
      
      clConf = analyseConflict(cl);

      backjumping(clConf);
      assert(tailTmp >= 0);

      tailTmp = tail;
      if(*clConf > 2)
	{
	  addClauseDB(clConf);	 	  
	  uncheckedEnqueue(clConf[1], clConf);
	
	  decrease();  
	  cla_inc *= clause_decay;
	}else if(*clConf == 2)
	{	  	
	  addClauseDB(clConf);	  
	  uncheckedEnqueue(clConf[1], binClause[clConf[2]].res);	  
	  free(clConf);	      	  
	}else
	{
	  uncheckedEnqueue(clConf[1], NULL);
	  free(clConf);
	}

      cl = propagate();
      if(cl) goto classicalBacktrack;      
    }
  
  while(tailTmp < tail)
    {      
      var = partial_assign[tailTmp++];      
      if(WASS_FALSE(var, atom[ABS(var)]))
	{
	  flipatom(var);
	  update_statistics_end_flip();
	}
    }
  update_statistics_end_flip();
}/* fixeVariable1 */


/**
   Hybride
 */
int pickHybride1(void)
{
  return pickcode[heuristic](fixeVariable1);
}/* pickHybride1 */


int indPrec = -1;

/**
   Update statistics end flip
 */
void update_statistics_end_flip(void)
{
  int i;

  if(maxSatNoBinaire > numfalse) maxSatNoBinaire = numfalse;

  if(maxSatTmp > (numfalse + numfalseBin))
    {
      maxSatTmp = numfalse + numfalseBin;
      if(maxSatTmp < maxSat) maxSat = maxSatTmp;
      for( i = 1 ; i <= numatom ; i++) bestInterpretation[i] = atom[i];
    }

  if((!numfalse) && (!numfalseBin))
    {
      for(i = 0 ; i < numclause ; i++) assert(numtruelit[i]);
      termineProg(SAT,"ASSIGNMENTS FOUND 1");
    } 
  assert(maxSatTmp >= maxSatNoBinaire);
  
}/* update_statistics_end_flip */


#ifndef NDEBUG
/**
   debug
 */
void debug()
{
  int mk[numatom + 1], bk[numatom + 1], fal = 0, falB = 0, numT;
  int i, j;
  unsigned int *cl, lit, var;
  
  for( i = 1 ; i <= numatom ; i++)
    {
      mk[i] = bk[i] = 0;
      if(isAssign[i] != NOT_ASS)
	assert((isAssign[i] % 2) == atom[i]);
    }

  for( i = 1 ; i<=*falseBin ; i++)
    {
      var = falseBin[i];
      assert(WASS_FALSE(var, atom[ABS(var)]));
    }

 /* les binaires */
  for( i = 1 ; i <= numatom ; i++)
    {
      if(WASS_TRUE(i<<1, atom[i]))
	lit = (i<<1) | 1;
      else lit = i<<1;
      
      for( j = 0 ; j<binClause[lit].nbBin ; j++)
	{	  
	  var = binClause[lit].tabBin[j];
	  if(WASS_FALSE(var, atom[ABS(var)]))
	    {
	      mk[ABS(var)]++;
	      falB++;
	    }
	  else bk[ABS(var)]++;
	}
    }

  for( i = 1 ; i <= numatom ; i++)
    {      
      assert(mk[i] == nbLitFalseBin[i]);
      if(mk[i]) assert(ABS(falseBin[whereFalseBin[i]]) == i);		
    }

  assert(numfalseBin == (falB>>1));

  for( i = 0 ; i<numclause ; i++)
    {
      cl = clause[i];
      
      numT = 0;
      for( j = 1 ; j <= *cl ; j++) if(WASS_TRUE(cl[j], atom[ABS(cl[j])])) numT++;
      assert(numtruelit[i] == numT);

      if(numT == 0)
	{
	  for( j = 1 ; j<=*cl ; j++) mk[ABS(cl[j])]++;
	  fal++;
	}

      if(numT == 1)
	{
	  for( j = 1 ; j<=*cl ; j++)
	    if(WASS_TRUE(cl[j], atom[ABS(cl[j])]))
	      {
		bk[ABS(cl[j])]++;
		break;
	      }
	}
    }
  
  assert(fal == numfalse);
 
  for( i = 1 ; i <= numatom ; i++) 
    {      
      assert(mk[i] == makecount[i]);
      if(!(bk[i] == breakcount[i]))
	printf("%d %d %d\n", i, bk[i], breakcount[i]);
      assert(bk[i] == breakcount[i]);
    }
}/* debug */
#endif


/**
   Fonction principal de WSAT
 */
void wsat()
{
  int unsigned var;
  int av;

  bornconf = 1;
  nbConflict = bornconf;
 
  /*********************************************************************************************/
  /*                                   Principal Loop                                          */
  /*********************************************************************************************/  
  numtry++;
  numflip = 0;
  init(0);
  
  lastNumflipRestart = 0;
  
  if((!numfalse) && (!numfalseBin)) termineProg(SAT,"ASSIGNMENTS FOUND");
  
  maxSatTmp = maxSat = numfalse + numfalseBin;
  
  numnullflip = 0;	
  
  /*********************************************************************************************/
  /*                                   Second Loop                                             */
  /*********************************************************************************************/
  while((numfalse > 0) || (numfalseBin > 0))
    {
      var = pickHybride1();
      
      if(var == NOVALUE) continue;	     		
      if(!decisionLevel) freeFlip++;

      if(changed[ABS(var)] == numflip) lastNumflipRestart -= nbRepetition++; else nbRepetition = 0;

      av = maxSatTmp;       
      flipatom(var);
      update_statistics_end_flip();
      if(av > maxSatTmp) lastNumflipRestart += av - maxSatTmp;

      if(nbConflict <= 0)
	{ 
	  cancelUntil(0);	  
	  nbRestartTotal++;
	  
	  nbConflict = (bornconf = super(nbRestartTotal) * INIT_LUBY); // luby64 strategie	  
	  lastNumflipRestart = nbConflict;
	  showInter();
	  nbConf = nbClauseAdd = 0;	  
	  numtry++;	  
	}

      if(numclauseLearnt >= sizeLearnt)	callCleaningDB();	 	
    }

  expertime = elapsed_seconds();
  numflip = 0;
  termineProg(OUT,"TIME_OUT");
}/* wsat */
